#include <allegro.h>
#include <string>
#include <sstream>
#include <ctime>
#include <fstream>
#include "main.h"
#include "keyStatesClass.h"
#include "Sound.h"
#include "iniReader.h"

using namespace std;

const int SOUNDS = 64;
#if 0
AGOGOHI  WAV       9,626      High Agogo
AGOGOLO  WAV      12,778      Low Agogo
BELLTREE WAV      65,948      Bell Tree
BONGOHI  WAV       9,778      High Bongo
BONGOLO  WAV       8,618      Low Bongo
CABASA   WAV       5,926      Cabasa
CASTINET WAV       3,382      Castinets
CLAPS    WAV      18,582      Handclap
CLAVES   WAV       4,686      Claves
CONGAHI1 WAV       5,198      Mute High Conga
CONGAHI2 WAV      19,860      Open High Conga
CONGALO  WAV      25,844      Low Conga
COWBELL  WAV      10,590      Cow Bell
CUICA1   WAV      18,326      Mute Cuica
CUICA2   WAV      25,458      Open Cuica
CYMBELL  WAV     159,978      Ride Bell
CYMCHINA WAV     193,330      Chinese Cymbal
CYMCRSH1 WAV      72,324      Crash Cymbal 1
CYMCRSH2 WAV      72,522      Crash Cymbal 2
CYMRIDE1 WAV      72,324      Ride Cymbal 1
CYMRIDE2 WAV      72,122      Ride Cymbal 2
CYMSPLSH WAV     139,120      Splash Cymbal
GUIRO1   WAV       4,048      Short Guiro
GUIRO2   WAV      26,198      Long Guiro
HIGHQ    WAV       2,820      High Q
HIHATCL  WAV       8,202      Closed High Hat
HIHATOP  WAV     164,168      Open High Hat
HIHATPD  WAV       6,626      Pedal High Hat
JINGLES  WAV      38,898      Jingle Bell
KICK1    WAV       8,122      Acoustic Bass
KICK2    WAV       9,494      Bass Drum 1
MARACAS  WAV       6,612      Maracas
METBELL  WAV      94,272      Metronome Bell
METCLICK WAV       6,874      Metal Click 
SCRATCH1 WAV      11,278      Scratch Push
SCRATCH2 WAV      16,808      Scratch Pull
SHAKER   WAV       3,392      Shaker
SLAP     WAV      19,692      Slap
SNARE1   WAV      24,426      Acoustic Snare
SNARE2   WAV      18,102      Electric Snare
SQRCLICK WAV         356      Square Click
STICKRIM WAV       6,178      Side Stick
STICKS   WAV      17,954      Sticks
SURDO1   WAV      18,816      Mute Surdo
SURDO2   WAV     146,152      Open Surdo
TAMBORIN WAV      15,400      Tambourine
TIMBALEH WAV      23,402      High Timbale
TIMBALEL WAV      29,640      Low Timbale
TOMHI1   WAV      42,346      High Tom
TOMHI2   WAV      42,740      High-Mid Tom
TOMLO1   WAV      69,368      High Floor Tom
TOMLO2   WAV      84,388      Low Floor Tom
TOMMID1  WAV      50,776      Low-Mid Tom
TOMMID2  WAV      63,582      Low Tom
TRIANGL1 WAV      12,070      Mute Triangle
TRIANGL2 WAV     152,888      Open Triangle
VIBSLAP  WAV      99,368      VibraSlap
WHISTLE1 WAV       8,210      Short Whistle
WHISTLE2 WAV      33,472      Long Whistle
WOODBLK1 WAV       7,792      High Wood Block
WOODBLK2 WAV      10,782      Low Wood Block
KICK64						  Bass drum 2
KICK3						  Electric bass drum
KICK5						  Bass drum 3
#endif
const char *fileNames[SOUNDS] = { "AGOGOHI", "AGOGOLO", "BELLTREE", "BONGOHI",
	"BONGOLO", "CABASA", "CASTINET", "CLAPS", "CLAVES", "CONGAHI1", "CONGAHI2",
	"CONGALO", "COWBELL", "CUICA1", "CUICA2", "CYMBELL", "CYMCHINA", "CYMCRSH1",
	"CYMCRSH2", "CYMRIDE1", "CYMRIDE2", "CYMSPLSH", "GUIRO1", "GUIRO2", "HIGHQ",
	"HIHATCL", "HIHATOP", "HIHATPD", "JINGLES", "KICK1", "KICK2", "MARACAS",
	"METBELL", "METCLICK", "SCRATCH1", "SCRATCH2", "SHAKER", "SLAP", "SNARE1",
	"SNARE2", "SQRCLICK", "STICKRIM", "STICKS", "SURDO1", "SURDO2", "TAMBORIN",
	"TIMBALEH", "TIMBALEL", "TOMHI1", "TOMHI2", "TOMLO1", "TOMLO2", "TOMMID1",
	"TOMMID2", "TRIANGL1", "TRIANGL2", "VIBSLAP", "WHISTLE1", "WHISTLE2",
	"WOODBLK1", "WOODBLK2", "KICK64", "KICK3", "KICK5" };

//timerCounterUpdater() will update timerCounter 32 times per second.
//This timer is used to make the game run at the same speed on all computers.
volatile int timerCounter = 1;
void timerCounterUpdater()
{
	++timerCounter;
} END_OF_STATIC_FUNCTION(timerCounterUpdater)

bool quitAll = false;
void quitAllProc()
{
	quitAll = true;
}

struct instrument
{
	Sound *sound;
	int soundOffset;
	int key;
	int fasterKey;
	int slowerKey;
	int stepSize;
	int autoPlayFreq;
	int oldClock;
	int volume;
	int pan;
	bool autoPlayActive;
};

instrument *loadDrumkit(const string path, int &instruments, Sound *sounds);

int main(int argc, char **argv)
{
	allegro_init();
	if(install_keyboard() < 0)
		return 0;
	if(install_mouse() < 0)
		return 0;
	enable_hardware_cursor();
	show_mouse(0);
	select_mouse_cursor(MOUSE_CURSOR_ARROW);
	show_mouse(screen);
	if(install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, NULL) < 0)
		return 0;
	int depth;
	if(!(depth = desktop_color_depth()))
		depth = 16;
	set_color_depth(depth);
	if(set_gfx_mode(GFX_AUTODETECT_WINDOWED, 320, 240, 0, 0) < 0)
		return 0;
	set_close_button_callback(quitAllProc);
	set_window_title("Master Lenman's Computer Drums");
	set_display_switch_mode(SWITCH_BACKGROUND);

    //Lock Timer Function and variable
	LOCK_FUNCTION(timerCounterUpdater);
	LOCK_VARIABLE(timerCounter);

	install_int_ex(timerCounterUpdater, BPS_TO_TIMER(1));
	
	BITMAP *cover = load_bitmap("cover.tga", 0);
	if(!cover)
		cover = create_bitmap(SCREEN_W, SCREEN_H);

	Sound sounds[SOUNDS];
	for(int i = 0; i < SOUNDS; ++i)
	{
		if(fileNames[i][0] == 0)
			continue;
		string path = string("sounds/") + fileNames[i] + ".WAV";
		if(!sounds[i].load(path))
			allegro_message((path + " could not be loaded!").c_str());
	}

	int instruments;
	instrument *instr = loadDrumkit("kits/default.ini", instruments, sounds);
	//We need this so multiple instruments can use the same key
	int keys[KEY_MAX];
	while(true)
	{
		if(keyStates[KEY_ESC])
		{
			string path = "kits/";
			if(fileSelect("Please select a drumkit file to load.", path, "ini", SCREEN_W, SCREEN_H) && fileExists(path))
			{
				instrument *instr_tmp = loadDrumkit(path, instruments, sounds);
				if(instr_tmp)
				{
					delete [] instr;
					instr = instr_tmp;
				}
				else
				{
					allegro_message("Error loading new drumkit!");
				}
			}
		}
		for(int i = 0; i < KEY_MAX; ++i)
			keys[i] = keyStates[i];
		while(timerCounter > 0)
		{
			blit(cover, screen, 0, 0, 0, 0, 320, 240);
			--timerCounter;
		}
		if(quitAll)
			break;
		if(!instr)
		{
			rest(1);
			continue;
		}
		int now = clock();
		for(int i = 0; i < instruments; ++i)
		{
			if(instr[i].autoPlayFreq != 0)
			{
				if(keys[instr[i].fasterKey] && instr[i].autoPlayFreq - instr[i].stepSize > 0)
					instr[i].autoPlayFreq -= instr[i].stepSize;
				if(keys[instr[i].slowerKey])
					instr[i].autoPlayFreq += instr[i].stepSize;
				if(keys[instr[i].key])
				{
					instr[i].autoPlayActive ^= 1;
					if(instr[i].autoPlayActive)
					{
						while(instr[i].oldClock < now)
							instr[i].oldClock += instr[i].autoPlayFreq;
					}
				}
				if(instr[i].autoPlayActive)
				{
					if(now >= instr[i].oldClock)
					{
						if(instr[i].sound)
							instr[i].sound->play(0, instr[i].volume, instr[i].pan);
						instr[i].oldClock += instr[i].autoPlayFreq;
					}
				}
			}
			else
			{
				if(keys[instr[i].key] && instr[i].sound)
					instr[i].sound->play(0, instr[i].volume, instr[i].pan);
			}
		}
		rest(1);
	}
	destroy_bitmap(cover);

	return 0;
} END_OF_MAIN()

int getKeyByName(const string &name)
{
	for(int j = 0; j < KEY_MAX; ++j)
	{
		if(name == keyNames[j])
		{
			return j;
		}
	}
	return 0;
}

instrument *loadDrumkit(const string path, int &instruments, Sound *sounds)
{
	iniReader kit;
	kit.load(path);
	instruments = kit.getIValue("instruments", 0);
	string name = kit.getValue("name", "");
	instrument *instr = 0;
	if(instruments > 0)
		instr = new instrument[instruments];
	else
		return instr;
	int now = clock();
	for(int i = 0; i < instruments; ++i)
	{
		kit.setCategory(string("instrument") + i2str(i + 1));
		string sound = kit.getValue("sound", "");
		instr[i].sound = 0;
		for(int j = 0; j < SOUNDS; ++j)
		{
			if(sound == fileNames[j])
			{
				instr[i].soundOffset = j;
				instr[i].sound = &(sounds[j]);
				break;
			}
		}
		string keyName = kit.getValue("key", "Esc");
		instr[i].key = getKeyByName(keyName);
		instr[i].autoPlayFreq = kit.getIValue("auto", 0);
		instr[i].oldClock = now + kit.getIValue("startauto", 0);
		instr[i].autoPlayActive = kit.getBValue("autoactive", true);
		keyName = kit.getValue("fasterkey", "Esc");
		instr[i].fasterKey = getKeyByName(keyName);
		keyName = kit.getValue("slowerkey", "Esc");
		instr[i].slowerKey = getKeyByName(keyName);
		instr[i].stepSize = kit.getIValue("stepsize", 0);
		
		instr[i].volume = kit.getIValue("volume", 255);
		instr[i].pan = kit.getIValue("pan", 127);
	}
	return instr;
}

string i2str(int i)
{
	ostringstream oss; // create a stream
	oss << i; // insert value to stream
	return oss.str(); // extract value and return it
}

string f2str(float i)
{
	ostringstream oss; // create a stream
	oss << i; // insert value to stream
	return oss.str(); // extract value and return it
}

//Turns a string into an integer, returns defaultValue if the string is no int value
int str2i(const std::string str, const int defaultValue)
{
	istringstream oss(str);
	int r;
	if(!(oss >> r))
		r = defaultValue;
	return r;
}

//Turns a string into an floating point number, returns defaultValue if the string is no
//floating point value
float str2f(const std::string str, const float defaultValue)
{
	istringstream oss(str);
	float r;
	if(!(oss >> r))
		r = defaultValue;
	return r;
}

bool fileSelect(std::string message, std::string &path, std::string ext, int w, int h)
{
	char pathBuffer[1024];
	strncpy(pathBuffer, path.c_str(), 1024);
	int retVal = file_select_ex(message.c_str(), pathBuffer, ext.c_str(), 1024, w, h);
	if(!retVal)
		return false;
	path = pathBuffer;
	return true;
}

bool fileExists(std::string filename)
{
	std::ifstream testStream(filename.c_str());
	bool result = !testStream.fail();
	testStream.close();

	return result;
}
